#!/usr/bin/env perl
# -*- perl -*-

#
# $Id: tksm,v 1.18 2001/10/13 21:43:14 eserte Exp $
# Author: Slaven Rezic
#
# Copyright (C) 1999, 2000, 2001 Slaven Rezic. All rights reserved.
# This program is free software; you can redistribute it and/or
# modify it under the same terms as Perl itself.
#
# Mail: eserte@cs.tu-berlin.de
# WWW:  http://user.cs.tu-berlin.de/~eserte/
#

=head1 NAME

tksm - multi-file search and replace

=head1 DESCRIPTION

This is a program for searching and replacing patterns in multiple files.

=cut

# XXX case insens: mit (?i) implementieren, checkbox

use Tk;
use Tk::ROText;
use Tk::Adjuster;
use Tk::Optionmenu;
use Tk::ErrorDialog;
use Tk::DropSite;
use Tk::DragDrop;
use File::Find;
use File::Basename;
use Cwd;
use strict;
use FindBin;
use lib "$FindBin::RealBin";

use vars qw($search $replace $directory $filemask $filemask_type $recursive
	    %ignore %ignore_label
	    $checked_files $found_files $found_matches $case_insens
	    @files $lb_i $act_file
	    $basemask $match_index
	    $top $VERSION $verbose
	    $find_timer $find_aborted);

$VERSION = sprintf("%d.%02d", q$Revision: 1.18 $ =~ /(\d+)\.(\d+)/);

$search = "";
$replace = "";
$directory = cwd;
$filemask  = ".*";
$filemask_type = "regexp";
$recursive = 1;
$ignore{'backups'}   = 1;
$ignore{'binaries'}  = 1;
$ignore{'vcs'}       = 1;
$ignore{'perl-blib'} = 1;
%ignore_label = ("backups" => "Backups",
		 "binaries" => "Binaries",
		 "vcs" => "CVS/RCS admin files",
		 "perl-blib" => "Perl build files (blib)",
		);
$checked_files = 0;
$found_files = 0;
$found_matches = 0;
$case_insens = 1;
$verbose = 1;

$top = new MainWindow;

if ($Tk::platform eq 'unix') {
    eval {
	require KDEUtil;
	my $kde = KDEUtil->new(-top => $top, -checkrunning => 1);
	$kde->kde_config_for_tk if $kde;
    }; warn $@ if $@;
}

use vars qw($readonly_bg $readonly_fg $normal_bg);
$readonly_bg = "#ffd0d0";
$readonly_fg = "#d04040";
$normal_bg;
if ($Tk::platform eq 'unix') {
    my $e = $top->Entry;
    my $lighter_bg = $normal_bg = $top->Darken($e->cget(-background), 120);
    $e->destroy;
    for (qw(Entry NumEntry PathEntry
	    Listbox KListbox K2Listbox TixHList HList Text
	    ROText BrowseEntry.Entry)) {
	$top->optionAdd("*$_.background", $lighter_bg, "userDefault");
    }
} else {
    $normal_bg = $top->Entry->cget(-background);
}

use vars qw($fixedfont);
$fixedfont = $top->optionGet("fixedFont", "FixedFont") || "Courier 10";

$top->title(basename($0) . " " . $VERSION);
my $mainf = $top->Frame->packAdjust(-fill => 'both',
				    -expand => 1,
				   );
use vars qw($ro);
$ro = $mainf->Scrolled('ROText',
		       -wrap => 'none',
		       -scrollbars => 'osoe',
		       -font => $fixedfont,
		      )->packAdjust(-side => 'left', -fill => 'both', -expand => 1);
$ro->tagConfigure('hilite', -background => 'red'); # XXX bold
$ro->tagConfigure('active', -background => 'red', -foreground => 'white'); # XXX bold
use vars qw($lb);
$lb = $mainf->Scrolled('Listbox',
		       -scrollbars => 'osoe',
		       -selectmode => 'single',
		      )->pack(-side => 'left',
			      -fill => 'both', -expand => 1);
$lb->bind('<ButtonPress-1>' => sub { set($lb->index('anchor')) });
$lb->bind('<3>' => sub {
    my $w = shift;
    my $e = $w->XEvent;
    my $y = $e->y;
    my $inx = $w->nearest($y);
    return if !defined $inx;
    my $file = $w->get($inx);
    my $ok = 0;
    if (is_in_path("emacsclient")) {
	system('emacsclient', '--no-wait', $file);
	if ($? == 0) {
	    $ok++;
	    # XXX or start emacs???
	}
    }
    if (!$ok) {
	if (defined $ENV{"EDITOR"} and
	    is_in_path($ENV{"EDITOR"}) and
	    is_in_path("xterm")) {
	    system("xterm -e $ENV{EDITOR} $file");
	} elsif ($^O eq 'MSWin32') {
	    if (is_in_path("ptked")) {
		system(1, "ptked", $file);
	    } else {
		system(1, "notepad", $file);
	    }
	} else {
	    die "Can't start editor.\n";
	}
    }
});

use vars qw($dnd);
$dnd = $lb->DragDrop
    (-startcommand => sub {
	 $dnd->configure(-text => basename(sel_file()));
	 0;
     },
     -sitetypes => ($^O eq 'MSWin32') ? ['Win32'] : ['Sun','XDND','KDE'],
     -event => '<B1-Motion>',
     -handlers => [[-type => 'FILE_NAME', \&dnd_handler]]);

my $ctrlf = $top->Frame->pack(-fill => 'x', -expand => 1);
$ctrlf->Button(-text => '<<',
	       -command => \&prev_file)->pack(-side => 'left');
$ctrlf->Button(-text => '<',
	       -command => \&prev_match)->pack(-side => 'left');
$ctrlf->Button(-text => 'Replace',
	       -command => \&do_replace)->pack(-side => 'left');
$ctrlf->Button(-text => 'Replace all',
	       -command => \&do_replace_all)->pack(-side => 'left');
$ctrlf->Button(-text => 'Replace all files',
	       -command => \&do_replace_all_files)->pack(-side => 'left');
$ctrlf->Button(-text => '>',
	       -command => \&next_match)->pack(-side => 'left');
$ctrlf->Button(-text => '>>',
	       -command => \&next_file)->pack(-side => 'left');

$ctrlf->Label(-textvariable => \$found_matches)->pack(-side => 'right');
$ctrlf->Label(-text => 'Matches:')->pack(-side => 'right');
$ctrlf->Label(-textvariable => \$found_files)->pack(-side => 'right');
$ctrlf->Label(-text => 'Files:')->pack(-side => 'right');
$ctrlf->Label(-textvariable => \$checked_files)->pack(-side => 'right');
$ctrlf->Label(-text => 'Checked:')->pack(-side => 'right');

my $optf = $top->Frame->pack(-fill => 'x', -expand => 1);
for (1..4) {
    $optf->gridColumnconfigure($_, -weight => 1);
}

my $row = 0;
my $col = 0;
$optf->Button(-text => 'Directory:',
	      -anchor => "w",
	      -command => \&set_dir,
	     )->grid(-row => $row, -column => $col++, -sticky => "we");
my $dir_e;
if (!eval '
    use Tk::PathEntry;
    $dir_e = $optf->PathEntry(-textvariable => \$directory);
    $dir_e->bind("<Return>" => sub { $dir_e->Finish });
    $dir_e->bind("<Escape>" => sub { $dir_e->Finish });
    1;
') {
    $dir_e = $optf->Entry(-textvariable => \$directory);
}
$dir_e->grid(-row => $row, -column => $col++,
	     -sticky => "ew",
	    );
if ($dir_e->can('DropSite')) {
    eval {
	$dir_e->DropSite
	    (-dropcommand => [\&accept_drop, $dir_e],
	     -droptypes => ($Tk::platform eq 'MSWin32' ? 'Win32' : ['KDE', 'XDND', 'Sun'])
	     );
    };
    warn __LINE__ . ": $@" if $@;
}

{
    my $om_dummy;
    $optf->Optionmenu(-options => [['Filemask (rx):' => 'regexp'],
				   ['Filemask (glob):' => 'glob'],
				  ],
		      -variable => \$filemask_type,
		      -textvariable => \$om_dummy,
		     )->grid(-row => $row, -column => $col++,
			     -sticky => "e");
}
$optf->Entry(-textvariable => \$filemask,
	     -width => 8,
	    )->grid(-row => $row, -column => $col++,
		    -sticky => "w",
		   );
$optf->Checkbutton(-text => 'Recursive',
		   -variable => \$recursive)->grid(-row => $row,
						   -column => $col++);
my $ign_mb = $optf->Menubutton(-text => "Ignore ...",
			       -relief => "raised",
			       -bd => 2,
			      )->grid(-row => $row, -column => $col++);
while(my($k) = each %ignore) {
    $ign_mb->checkbutton(-label => $ignore_label{$k},
			 -variable => \$ignore{$k});
}
$row++;

my $colspan = 2;
$optf->Button(-text => 'Search:',
	      -anchor => "w",
	      -command => \&search,
	     )->grid(-row => $row, -column => 0, -sticky => "we");
my $se =
    $optf->Entry(-textvariable => \$search)->grid(-row => $row, -column => 1,
						  -columnspan => $colspan,
						  -sticky => "ew",
						 );
$se->focus;
$row++;

$optf->Label(-text => 'Replace:',
	     -anchor => "w",
	     -borderwidth => 3,
	     -padx => 11,
	    )->grid(-row => $row, -column => 0, -sticky => "we");
my $re =
    $optf->Entry(-textvariable => \$replace)->grid(-row => $row, -column => 1,
						   -columnspan => $colspan,
						   -sticky => "ew",
						  );

$se->bind("<Return>" => sub { search(); $re->focus });

# XXX aktivieren!
#  $optf->Checkbutton
#      (-text => 'Case insens',
#       -variable => \$case_insens)->grid(-row => $row, -column => 2);
$row++;


MainLoop;

# RELOADER_START

sub sel_file { $lb->get($lb->curselection) }

sub dnd_handler {
    my ($offset,$max) = @_;
    my $file = sel_file();
    $file;
}

sub set_dir {
#      require Tk::FileSelect;
#      my $dir = $top->FileSelect(-verify => [qw(-d)])->Show;
#      if ($dir ne "") {
#  	$directory = $dir;
#      }

    my $dir = choose_dir($top);
    if (defined $dir) {
	$directory = $dir;
    }
}

sub prev_match {
    my $len;
    my $new_index = $ro->search(qw(-backwards -regexp -nocase --), $search,
				$match_index, 'end');
    if (!defined $new_index || $new_index eq '') {
	prev_file();
    } else {
	$match_index = $new_index;
	$ro->see($match_index);
    }
}

sub next_match {
    if (!@files) {
	get_files();
    }
    my $len;
warn "newi=$match_index";
    my $new_index = $ro->search(-count => \$len,
				qw(-regexp -nocase --), $search,
				$match_index, 'end');
    my(@ranges) = $ro->tagRanges('active');
    if (@ranges) {
	$ro->tagRemove('active', '1.0', 'end');
	$ro->tagAdd('hilite', @ranges);
    }
    if (!defined $new_index || $new_index eq '') {
	next_file();
    } else {
#XXXX SEGFAULT! (?)
warn "inx = <$new_index + $len c>";
	$ro->tagAdd('active', $new_index, "$new_index + $len c");
warn "newi=$new_index";
	$match_index = $ro->index("$new_index + 1 c");
	$ro->see($match_index);
    }
}


sub prev_file {
    $lb_i--;
    if ($lb_i >= 0) {
	set();
	1;
    } else {
	empty();
	0;
    }
}

sub next_file {
    $lb_i++;
    if ($lb_i <= $lb->size-1) {
	set();
	1;
    } else {
	empty();
	0;
    }
}

sub set {
    if (@_) {
	$lb_i = $_[0];
    }
    $lb->selectionClear(0, 'end');
    $lb->selectionSet($lb_i);
    $lb->see($lb_i);
    $act_file = $lb->get($lb_i);
    show_file();
}

sub empty {
    undef $act_file;
    $lb->selectionClear(0, 'end');
    $ro->delete('1.0', 'end');
}

sub show_file {
    return if $act_file eq '';
    $ro->delete('1.0', 'end');
    if (open(F, $act_file)) {
	binmode F;
	my $l = 1;
	my $first = 1;
	while(<F>) {
	    $ro->insert("$l.0", $_);
	    if (/$search/i) {
		if ($first) {
		    $ro->see("$l.0");
		    $first = 0;
		    $match_index = "$l." . int(length($`) + length($&));
		}
		while (/$search/ig) {
		    my $pos = pos();
		    if (defined $pos) {
			$ro->tagAdd('hilite',
				    "$l." . ($pos-length($&)),
				    "$l." . $pos);
		    }
		}
	    }
	    $l++;
	}
	close F;

	if (-w $act_file) {
	    $ro->configure(-bg => $normal_bg);
	} else {
	    $ro->configure(-bg => $readonly_bg);
	}
    }
    if (!defined $match_index) { $match_index = '1.0' }
}

sub do_replace {
    die "NYI";
}

sub do_replace_all {
    my $nonext = shift;
    return if $act_file eq '';
    if (open(R, $act_file)) {
	binmode R;
	if (open(W, ">$act_file~")) {
	    binmode W;
	    my $err;
	    while(<R>) {
		eval {
		    if ($case_insens) {
			s/\Q$search\E/$replace/gi;
		    } else {
			s/\Q$search\E/$replace/g;
		    }
		}; if ($@) { $err = $@ }
		print W $_;
	    }
	    if (defined $err) {
		warn "Error while substituting $act_file:\n$err";
	    }
	    close W;
	    copy_stat($act_file, "$act_file~", -notime => 1);
	    close R;
	    warn "Try to unlink $act_file\n" if $verbose;
	    if (!unlink $act_file) {
		warn "Couldn't unlink $act_file: $!\n" if $verbose;
	    } else {
		warn "Try to rename $act_file~ to $act_file\n"
		    if $verbose;
		if (!rename "$act_file~", $act_file) {
		    warn "Couldn't rename $act_file~: $!\n" if $verbose;
		}
	    }
	} else {
	    close R;
	}
    }
    next_file() unless $nonext;
}

sub do_replace_all_files {
    while(next_file()) {
	do_replace_all('nonext');
	$top->idletasks;
    }
}

sub search {
    get_files();
}

sub get_files {

    @files = ();
    return if $directory eq '';
    return if $search eq '' or $search eq '.';
    my $topdir = $directory;

    if ($filemask eq '') {
	$filemask = ($filemask_type eq 'regexp' ? '.*' : '*');
    }
    if ($filemask_type eq 'regexp') {
	$basemask = $filemask;
    } else {
	$basemask = $filemask;
	$basemask =~ s/\./\\./g;
	$basemask =~ s/\*/.*/g;
	$basemask =~ s/\?/./g;
	# other to follow...
    }

    $checked_files = 0;
    $found_files = 0;
    $found_matches = 0;
    $lb->delete(0, 'end');
    $lb->xview(0);
    empty();

    if (!-d $topdir) {
	die "$topdir is no directory";
    }

    # prepare find
    my $io;
    if ($Tk::platform eq 'unix') {
	require Tk::InputO;
	$io = $top->InputO(-cursor => "watch")->place(-relx => 0, -rely => 0,
						      -relheight => 1,
						      -relwidth => 1);
	$io->bind("all", "<Escape>" => sub { $find_aborted = 1 });
    } else {
	$top->Busy;
    }
    $find_timer = Tk::timeofday;
    undef $find_aborted;

    # do find
    eval {
	find(\&wanted, $topdir);
    };

    # unprepare find
    if ($io) {
	my $err = $@;
	$io->destroy;
	die "$err\n" if $err;
    } else {
	warn $@ if $@;
	$top->Unbusy;
    }

    eval {
	my $w = $lb->fontMeasure($lb->cget(-font), $directory);
	my $w0 = $lb->fontMeasure($lb->cget(-font), "0");
	$lb->xview(int($w/$w0));
    };
    warn $@ if $@;

    $lb_i = -1;
    if (@files) {
	next_file();
    }
}

sub wanted {
    # Escape watcher
    if (Tk::timeofday()-$find_timer>0.2) {
	$top->update;
	$find_timer = Tk::timeofday();
    }
    if ($find_aborted) {
	die "Aborted by user";
    }

    # ignore ...
    return if $_ eq '.';
    return if ($ignore{'backups'} && /(\.bak|~)$/i);
    return if ($ignore{'vcs'} &&
	       $File::Find::name =~ m;(/CVS/.*|/RCS/.*,v|\.prj)$;);
    return if ($ignore{'perl-blib'} &&
	       ($File::Find::dir =~ m|/blib/| || $_ eq 'Makefile.old'));

    # grep
    if (-f $_ && /^$basemask$/) {
	if (open(F, $_)) {
	    binmode F;
	    my $l;
	    my $found = 0;
	    while($l = scalar <F>) {
		return if ($ignore{'binaries'} && $l =~ /[\200-\237]/);
		if ($l =~ /$search/i) { # case ? XXX
		    $found++;
		}
	    }
	    close F;
	    $checked_files++;
	    if ($found) {
		push @files, $File::Find::name;
		$found_files++;
		$found_matches += $found;
		$lb->insert('end', $File::Find::name);
	    }
	}
    }
    if (-d $_ && !$recursive) {
	$File::Find::prune = 1;
    }
}

# REPO BEGIN
# REPO MD5 a77759517bc00f13c52bb91d861d07d0

=head2 file_name_is_absolute($file)

=for category File

Return true, if supplied file name is absolute. This is only necessary
for older perls where File::Spec is not part of the system.

=cut

sub file_name_is_absolute {
    my $file = shift;
    my $r;
    eval {
        require File::Spec;
        $r = File::Spec->file_name_is_absolute($file);
    };
    if ($@) {
	if ($^O eq 'MSWin32') {
	    $r = ($file =~ m;^([a-z]:(/|\\)|\\\\|//);i);
	} else {
	    $r = ($file =~ m|^/|);
	}
    }
    $r;
}
# REPO END

# REPO BEGIN
# REPO MD5 1b42243230d92021e6c361e37c9771d1

=head2 is_in_path($prog)

=for category File

Return the pathname of $prog, if the program is in the PATH, or undef
otherwise.

DEPENDENCY: file_name_is_absolute

=cut

sub is_in_path {
    my($prog) = @_;
    return $prog if (file_name_is_absolute($prog) and -f $prog and -x $prog);
    require Config;
    my $sep = $Config::Config{'path_sep'} || ':';
    foreach (split(/$sep/o, $ENV{PATH})) {
	if ($^O eq 'MSWin32') {
	    return "$_\\$prog"
		if (-x "$_\\$prog.bat" ||
		    -x "$_\\$prog.com" ||
		    -x "$_\\$prog.exe");
	} else {
	    return "$_/$prog" if (-x "$_/$prog");
	}
    }
    undef;
}
# REPO END

# REPO BEGIN
# REPO MD5 f567def1f7ce8f3361e474b026594660

=head2 copy_stat($src, $dest, %args)

=for category File

Copy stat information (owner, group, mode and time) from one file to
another. If $src is an array reference, then this is used as the
source stat information. With %args some aspects can be ignored
(-nomode, -noowner, -notime).

=cut

sub copy_stat {
    my($src, $dest, %args) = @_;
    my @stat = ref $src eq 'ARRAY' ? @$src : stat($src);
    die "Can't stat $src: $!" if !@stat;

    if (!$args{-nomode}) {
	chmod $stat[2], $dest
	    or warn "Can't chmod $dest to " . sprintf("0%o", $stat[2]) . ": $!";
    }
    if (!$args{-noowner}) {
	chown $stat[4], $stat[5], $dest
	    or do {
		my $save_err = $!; # otherwise it's lost in the get... calls
		warn "Can't chown $dest to " .
		    (getpwuid($stat[4]))[0] . "/" .
		    (getgrgid($stat[5]))[0] . ": $save_err";
	    };
    }
    if (!$args{-notime}) {
	utime $stat[8], $stat[9], $dest
	    or warn "Can't utime $dest to " .
		scalar(localtime $stat[8]) . "/" .
		scalar(localtime $stat[9]) .
		": $!";
    }
}
# REPO END

# from PerlMonth
sub choose_dir {
    my $top = shift;

    my $curr_dir;
    if (defined $::choose_dir_curr_dir &&
	-d $::choose_dir_curr_dir) {
	$curr_dir = $::choose_dir_curr_dir;
    } else {
	$curr_dir = Cwd::cwd();
    }

    eval {
	require Tk::DirSelect;
	$curr_dir = $top->DirSelect(-dir => $curr_dir)->Show;
    };
    if (!$@) {
	require File::Spec;
	$curr_dir = File::Spec->canonpath($curr_dir);
	$::choose_dir_curr_dir = $curr_dir;
	return $curr_dir;
    } else {
	warn $@ if $@;
    }

    require Tk::DirTree;

    my $t = $top->Toplevel;
    $t->title("Choose directory:");
    my $ok = 0; # flag: "1" means OK, "-1" means cancelled

    # Create Frame widget before the DirTree widget, so it's always visible
    # if the window gets resized.
    my $f = $t->Frame->pack(-fill => "x", -side => "bottom");

    my $d;
    $d = $t->Scrolled('DirTree',
		      -scrollbars => 'osoe',
		      -width => 35,
		      -height => 20,
		      -selectmode => 'browse',
		      -exportselection => 1,
		      -browsecmd => sub { $curr_dir = shift },

		      # With this version of -command a double-click will
		      # select the directory
		      -command   => sub { $ok = 1 },

		      # With this version of -command a double-click will
		      # open a directory. Selection is only possible with
		      # the Ok button.
		      #-command   => sub { $d->opencmd($_[0]) },
		     )->pack(-fill => "both", -expand => 1);
    # Set the initial directory
    $d->chdir($curr_dir);

    $f->Button(-text => 'Ok',
	       -command => sub { $ok =  1 })->pack(-side => 'left');
    $f->Button(-text => 'Cancel',
	       -command => sub { $ok = -1 })->pack(-side => 'left');

    # You probably want to set a grab. See the Tk::FBox source code for
    # more information (search for grabCurrent, waitVariable and
    # grabRelease).
    $f->waitVariable(\$ok);
    $t->destroy;

    if ($ok == 1) {
	require File::Spec;
	$curr_dir = File::Spec->canonpath($curr_dir);
	$::choose_dir_curr_dir = $curr_dir;
	return $curr_dir;
    }

    undef;
}

sub accept_drop {
    my($w, $seln) = @_;
    my $filename;
    my @targ = $w->SelectionGet('-selection'=>$seln,'TARGETS');
    foreach (@targ) {
	if (/FILE_NAME/) {
	    $filename = $w->SelectionGet('-selection'=>$seln,'FILE_NAME');
	    last;
	}
	if ($Tk::platform eq 'MSWin32' && /STRING/) {
	    $filename = $w->SelectionGet('-selection'=>$seln,$_);
	    last;
	}
    }
    if (defined $filename && -d $filename) {
	$directory = $filename;
    }
}

__END__

=head1 AUTHOR

Slaven Rezic <eserte@onlineoffice.de>

=cut
